from evdev import InputDevice, ecodes  # type: ignore
import threading

from bluer_ugv.swallow.session.classical.leds import ClassicalLeds
from bluer_ugv.swallow.session.classical.setpoint.classes import ClassicalSetPoint
from bluer_ugv.logger import logger


class ClassicalMousePad:
    def __init__(
        self,
        leds: ClassicalLeds,
        setpoint: ClassicalSetPoint,
    ):
        self.leds = leds
        self.setpoint = setpoint

        try:
            self.device = InputDevice("/dev/input/event0")
            logger.info(
                "{}: using {}.".format(
                    self.__class__.__name__,
                    self.device.name,
                )
            )
        except Exception as e:
            logger.warning(e)
            self.device = None
            return

        self._thread = threading.Thread(
            target=self.run_,
            daemon=True,
        )
        self._thread.start()

    def run_(self) -> bool:
        logger.info(f"{self.__class__.__name__}: thread started.")

        for event in self.device.read_loop():
            if event.type == ecodes.EV_REL:
                if event.code == ecodes.REL_Y and self.setpoint.started:
                    self.setpoint.put(
                        what="speed",
                        value=self.setpoint.get(what="speed") - event.value,
                    )  # up/down
                elif event.code == ecodes.REL_X:
                    self.setpoint.put(
                        what="steering",
                        value=-event.value,
                    )  # left/right

                self.leds.flash("yellow")

            elif (
                event.type == ecodes.EV_KEY
                and event.code == ecodes.BTN_LEFT
                and event.value == 0
            ):
                self.setpoint.stop()

        return True



from bluer_objects.README.items import ImageItems

from bluer_ugv.README.consts import bluer_ugv_assets
from bluer_ugv.README.arzhang.consts import arzhang_assets2
from bluer_ugv.README.rangin.consts import rangin_assets2
from bluer_ugv.README.swallow.consts import swallow_assets2


docs = [
    {
        "path": "../docs/UGVs",
    },
    {
        "path": "../docs/UGVs/swallow.md",
        "items": ImageItems(
            {
                f"{swallow_assets2}/20250701_2206342_1.gif": "",
                f"{swallow_assets2}/20250913_203635~2_1.gif": "",
            }
        ),
    },
    {
        "path": "../docs/UGVs/arzhang.md",
        "items": ImageItems(
            {
                f"{arzhang_assets2}/20251005_112530.jpg": "",
            }
        ),
    },
    {
        "path": "../docs/UGVs/arzhang2.md",
        "items": ImageItems(
            {
                f"{arzhang_assets2}/20251005_112530.jpg": "",
            }
        ),
    },
    {
        "path": "../docs/UGVs/arzhang3.md",
        "items": ImageItems({f"{bluer_ugv_assets}/bluer-light.png": ""}),
    },
    {
        "path": "../docs/UGVs/rangin.md",
        "items": ImageItems({f"{rangin_assets2}/rangin.png": ""}),
    },
]



from bluer_objects.README.items import ImageItems
from bluer_sbc.parts.db import db_of_parts
from bluer_sbc.parts.consts import parts_url_prefix

from bluer_ugv.README.swallow.consts import (
    swallow_assets2,
    swallow_electrical_designs,
)
from bluer_ugv.designs.swallow.parts import dict_of_parts
from bluer_ugv.README.swallow.digital.design import mechanical, ultrasonic_sensor


docs = (
    [
        {
            "path": "../docs/swallow/digital/design",
        },
        {
            "path": "../docs/swallow/digital/design/computers.md",
        },
        {
            "path": "../docs/swallow/digital/design/operation.md",
            "items": ImageItems(
                {
                    f"{swallow_assets2}/20251005_113232.jpg": "",
                }
            ),
        },
        {
            "path": "../docs/swallow/digital/design/parts.md",
            "items": db_of_parts.as_images(
                dict_of_parts,
                reference=parts_url_prefix,
            ),
            "macros": {
                "parts:::": db_of_parts.as_list(
                    dict_of_parts,
                    reference=parts_url_prefix,
                    log=False,
                ),
            },
        },
        {
            "path": "../docs/swallow/digital/design/terraform.md",
            "items": ImageItems(
                {
                    f"{swallow_assets2}/20250611_100917.jpg": "",
                    f"{swallow_assets2}/lab.png": "",
                    f"{swallow_assets2}/lab2.png": "",
                }
            ),
        },
        {
            "path": "../docs/swallow/digital/design/steering-over-current-detection.md",
            "items": ImageItems(
                {
                    f"{swallow_electrical_designs}/steering-over-current.png": f"{swallow_electrical_designs}/steering-over-current.svg",
                }
            ),
        },
        {
            "path": "../docs/swallow/digital/design/rpi-pinout.md",
        },
    ]
    + mechanical.docs
    + ultrasonic_sensor.docs
)



import threading
from typing import Union, Dict

from bluer_ugv.swallow.session.classical.leds import ClassicalLeds
from bluer_ugv.swallow.session.classical.setpoint.steering import (
    generate_left_and_right,
)
from bluer_ugv.logger import logger


class ClassicalSetPoint:
    def __init__(
        self,
        leds: ClassicalLeds,
    ):
        self.speed = 0
        self.steering = 0
        self.started = False

        self.leds = leds

        self._lock = threading.Lock()

    def get(
        self,
        what: str = "all",
    ) -> Union[int, bool, Dict[str, Union[int, bool]]]:
        with self._lock:
            if what == "all":
                return {
                    "speed": self.speed,
                    "started": self.started,
                    "steering": self.steering,
                }

            if what == "left":
                return generate_left_and_right(self.speed, self.steering)[0]

            if what == "right":
                return generate_left_and_right(self.speed, self.steering)[1]

            if what == "speed":
                return self.speed

            if what == "started":
                return self.started

            if what == "steering":
                return self.steering

            logger.error(f"{self.__class__.__name__}.get: {what} not found.")
            return 0

    def put(
        self,
        value: Union[int, bool, Dict[str, Union[int, bool]]],
        what: str = "all",
        log: bool = False,
    ):
        with self._lock:
            if what == "all":
                self.speed = min(100, max(-100, int(value["speed"])))
                self.started = bool(value["started"])
                self.steering = min(100, max(-100, int(value["steering"])))
                return

            if what == "speed":
                self.speed = min(100, max(-100, int(value)))
                if log:
                    logger.info(
                        "{}.put: speed={}".format(
                            self.__class__.__name__,
                            self.speed,
                        )
                    )
                return

            if what == "started":
                self.started = bool(value)
                if log:
                    logger.info(
                        "{}.put: {}".format(
                            self.__class__.__name__,
                            "started" if value else "stopped",
                        )
                    )
                return

            if what == "steering":
                self.steering = min(100, max(-100, int(value)))
                if log:
                    logger.info(
                        "{}.put: steering={}".format(
                            self.__class__.__name__,
                            self.steering,
                        )
                    )
                return

            logger.error(f"{self.__class__.__name__}.put: {what} not found.")

    def start(self):
        self.put(
            {
                "speed": 0,
                "started": True,
                "steering": 0,
            }
        )

        logger.info(f"{self.__class__.__name__}.start")

    def stop(self):
        self.put(
            {
                "speed": 0,
                "started": False,
                "steering": 0,
            }
        )

        logger.info(f"{self.__class__.__name__}.stop")

        self.leds.set("red", False)
        self.leds.set("yellow", False)

    def update(self) -> bool:
        with self._lock:
            if self.started:
                self.leds.flash("red")

        return True



from bluer_objects.README.items import ImageItems, Items
from bluer_sbc.parts.db import db_of_parts
from bluer_sbc.parts.consts import parts_url_prefix

from bluer_ugv.README.rangin.items import items
from bluer_ugv.designs.rangin.parts import dict_of_parts
from bluer_ugv.README.rangin.consts import rangin_mechanical_design

docs = [
    {
        "items": items,
        "path": "../docs/rangin",
    },
    {
        "path": "../docs/rangin/parts.md",
        "items": db_of_parts.as_images(
            dict_of_parts,
            reference=parts_url_prefix,
        ),
        "macros": {
            "parts:::": db_of_parts.as_list(
                dict_of_parts,
                reference=parts_url_prefix,
                log=False,
            ),
        },
    },
    {
        "path": "../docs/rangin/mechanical.md",
        "items": Items(
            [
                {
                    "name": "arzhang",
                    "marquee": f"{rangin_mechanical_design}/robot.png?raw=true",
                    "url": f"{rangin_mechanical_design}/robot.stl",
                },
                {
                    "name": "90",
                    "marquee": f"{rangin_mechanical_design}/robot-90.png?raw=true",
                    "url": f"{rangin_mechanical_design}/robot-90.stl",
                },
                {
                    "name": "90 (without the cage)",
                    "marquee": f"{rangin_mechanical_design}/robot-90-2.png?raw=true",
                    "url": f"{rangin_mechanical_design}/robot-90.stl",
                },
            ]
        ),
    },
    {
        "path": "../docs/rangin/computers.md",
    },
]



from bluer_ugv.swallow.session.classical.motor.left import ClassicalLeftMotor
from bluer_ugv.swallow.session.classical.motor.right import ClassicalRightMotor
from bluer_ugv.swallow.session.classical.motor.rear import ClassicalRearMotors
from bluer_ugv.swallow.session.classical.motor.steering import ClassicalSteeringMotor



from bluer_objects.README.consts import assets_url, github_kamangir


algo_docs = f"{github_kamangir}/bluer-algo/blob/main/bluer_algo/docs"

bluer_ugv_assets = assets_url("bluer-ugv")

bluer_ugv_assets2 = assets_url(
    suffix="bluer-ugv",
    volume=2,
)



dict_of_parts = {
    "4-ch-transceiver": "",
    "2xAA-battery-holder": "",
    "4xAA-battery-holder": "",
    "yellow-gearbox-dc-motor": "4 x 270 RPM",
    "yellow-wheels": "x 4",
    "L-1x2": "x 8",
    "M3": "🚧",
}



from blueness import module

from bluer_ugv import NAME
from bluer_ugv.swallow.session.classical.ultrasonic_sensor.log import (
    UltrasonicSensorDetectionLog,
)
from bluer_ugv.logger import logger


NAME = module.name(__file__, NAME)


def review(
    object_name: str,
    export_gif: bool = False,
    frame_count: int = -1,
    log: bool = True,
    rm_blank: bool = True,
) -> bool:
    logger.info(
        "{}.review{}: {}".format(
            NAME,
            f" [{frame_count} frame(s)]" if frame_count == -1 else "",
            object_name,
        )
    )

    detection_log = UltrasonicSensorDetectionLog()

    if not detection_log.load(
        object_name=object_name,
    ):
        return False

    if not detection_log.export(
        object_name=object_name,
        line_width=80,
        export_gif=export_gif,
        frame_count=frame_count,
        log=log,
        rm_blank=rm_blank,
    ):
        return False

    return True



from typing import Tuple
import numpy as np

from bluer_options.timer import Timer
from bluer_options import string
from bluer_algo.tracker.classes.target import Target
from bluer_algo.tracker.classes.camshift import CamShiftTracker
from bluer_sbc.imager.camera import instance as camera

from bluer_ugv import env
from bluer_ugv.swallow.session.classical.camera.generic import ClassicalCamera
from bluer_ugv.swallow.session.classical.camera.generic import ClassicalCamera
from bluer_ugv.swallow.session.classical.keyboard import ClassicalKeyboard
from bluer_ugv.swallow.session.classical.leds import ClassicalLeds
from bluer_ugv.swallow.session.classical.setpoint.classes import ClassicalSetPoint
from bluer_ugv.swallow.session.classical.mode import OperationMode
from bluer_ugv.logger import logger


class ClassicalTrackingCamera(ClassicalCamera):
    def __init__(
        self,
        keyboard: ClassicalKeyboard,
        leds: ClassicalLeds,
        setpoint: ClassicalSetPoint,
        object_name: str,
    ):
        super().__init__(keyboard, leds, setpoint, object_name)

        self.track_window: Tuple[int, int, int, int] = None

        self.tracking_timer = Timer(
            period=env.BLUER_UGV_CAMERA_ACTION_PERIOD,
            name="{}.tracking".format(self.__class__.__name__),
            log=True,
        )

        self.tracker = CamShiftTracker()

    def initialize(self) -> bool:
        if not super().initialize():
            return False

        return self.select_target()

    def select_target(self) -> bool:
        success, image = camera.capture(
            close_after=False,
            open_before=False,
            log=True,
        )
        if not success:
            return success

        self.leds.set_all(True)
        success, self.track_window = Target.select(
            np.flip(image, axis=2),
            local=False,
        )
        self.leds.set_all(False)
        if not success:
            return success

        logger.info(f"track_window: {self.track_window}")

        self.tracker.start(
            frame=image,
            track_window=self.track_window,
        )

        return True

    def update(self) -> bool:
        if not super().update():
            return False

        mode = self.keyboard.get("mode", OperationMode.NONE)
        if mode == OperationMode.TRAINING:
            return self.update_training()

        if self.setpoint.speed <= 0:
            return True

        if mode == OperationMode.ACTION:
            return self.update_action()

        return True

    def update_action(self) -> bool:
        if not self.tracking_timer.tick():
            return True

        self.leds.flash("red")

        success, image = camera.capture(
            close_after=False,
            open_before=False,
            log=True,
        )
        if not success:
            return success

        _, self.track_window, _ = self.tracker.track(
            frame=image,
            track_window=self.track_window,
        )

        x, _, w, _ = self.track_window
        if x + w // 2 > image.shape[1] * 2 / 3:
            self.setpoint.put(
                what="steering",
                value=-env.BLUER_UGV_SWALLOW_STEERING_SETPOINT,
                log=True,
            )
        elif x + w // 2 < image.shape[1] * 1 / 3:
            self.setpoint.put(
                what="steering",
                value=env.BLUER_UGV_SWALLOW_STEERING_SETPOINT,
                log=True,
            )

        return True

    def update_training(self) -> bool:
        self.keyboardset("mode", OperationMode.NONE)
        return self.select_target()



from bluer_objects.README.items import ImageItems
from bluer_objects.README.consts import assets2

from bluer_ugv.README.swallow.consts import (
    swallow_assets2,
    swallow_ultrasonic_sensor_designs,
)

docs = [
    {
        "path": "../docs/swallow/digital/design/ultrasonic-sensor",
        "items": ImageItems(
            {
                f"{swallow_ultrasonic_sensor_designs}/geometry.png?raw=true": f"{swallow_ultrasonic_sensor_designs}/geometry.svg",
            }
        ),
    },
    {
        "path": "../docs/swallow/digital/design/ultrasonic-sensor/dev.md",
        "cols": 1,
        "items": ImageItems(
            {
                f"{swallow_assets2}/20251001_203056_1.gif": "",
                f"{swallow_assets2}/20251001_185852.jpg": "",
            }
        ),
    },
    {
        "path": "../docs/swallow/digital/design/ultrasonic-sensor/tester.md",
        "cols": 2,
        "items": ImageItems(
            {
                f"{swallow_assets2}/20250918_122725.jpg": "",
                f"{swallow_assets2}/20250918_194715-2.jpg": "",
                f"{swallow_assets2}/20250918_194804_1.gif": "",
                f"{assets2}/ultrasonic-sensor-tester/00.jpg": "",
            }
        ),
    },
    {
        "path": "../docs/swallow/digital/design/ultrasonic-sensor/shield.md",
        "items": ImageItems(
            {
                f"{swallow_assets2}/20250923_142200.jpg": "",
                f"{swallow_assets2}/20250923_145111.jpg": "",
            }
        ),
    },
]



import os
from bluer_options.env import load_config, load_env, get_env

load_env(__name__)
load_config(__name__)


BLUER_UGV_BEAST_MODEL = get_env(
    "BLUER_UGV_BEAST_MODEL",
)

BLUER_UGV_CAMERA_TRAINING_PERIOD = get_env(
    "BLUER_UGV_CAMERA_TRAINING_PERIOD",
    3.0,
)
BLUER_UGV_CAMERA_ACTION_PERIOD = get_env(
    "BLUER_UGV_CAMERA_ACTION_PERIOD",
    1.5,
)

BLUER_UGV_SWALLOW_STEERING_SETPOINT = get_env(
    "BLUER_UGV_SWALLOW_STEERING_SETPOINT",
    50,
)

BLUER_UGV_MOUSEPAD_ENABLED = get_env(
    "BLUER_UGV_MOUSEPAD_ENABLED",
    0,
)

BLUER_UGV_SWALLOW_NAVIGATION_DATASET_LIST = get_env(
    "BLUER_UGV_SWALLOW_NAVIGATION_DATASET_LIST"
)

BLUER_UGV_SWALLOW_NAVIGATION_MODEL = get_env(
    "BLUER_UGV_SWALLOW_NAVIGATION_MODEL",
)

BLUER_UGV_SWALLOW_YOLO_MODEL = get_env(
    "BLUER_UGV_SWALLOW_YOLO_MODEL",
)

BLUER_UGV_SWALLOW_YOLO_DATASET_LIST = get_env(
    "BLUER_UGV_SWALLOW_YOLO_DATASET_LIST",
)

BLUER_UGV_ULTRASONIC_SENSOR_TEST_OBJECT = get_env(
    "BLUER_UGV_ULTRASONIC_SENSOR_TEST_OBJECT"
)

BLUER_UGV_ULTRASONIC_SENSOR_WARNING_THRESHOLD = get_env(
    "BLUER_UGV_ULTRASONIC_SENSOR_WARNING_THRESHOLD",
    800.0,
)
BLUER_UGV_ULTRASONIC_SENSOR_DANGER_THRESHOLD = get_env(
    "BLUER_UGV_ULTRASONIC_SENSOR_DANGER_THRESHOLD",
    500.0,
)

BLUER_UGV_ULTRASONIC_SENSOR_ENABLED = get_env(
    "BLUER_UGV_ULTRASONIC_SENSOR_ENABLED",
    1,
)

BLUER_UGV_ULTRASONIC_SENSOR_MAX_M = get_env(
    "BLUER_UGV_ULTRASONIC_SENSOR_",
    0.80,
)

BLUER_UGV_ULTRASONIC_SENSOR_KEEP_LOG = get_env(
    "BLUER_UGV_ULTRASONIC_SENSOR_KEEP_LOG",
    1,
)

BLUER_UGV_ULTRASONIC_SENSOR_LOG = get_env(
    "BLUER_UGV_ULTRASONIC_SENSOR_LOG",
    1,
)



dict_of_parts = {
    "4-ch-transceiver": "",
    "2xAA-battery-holder": "",
    "4xAA-battery-holder": "",
    "yellow-gearbox-dc-motor": "2 x 270 RPM + 1 x 80 RPM",
    "yellow-wheels": "x 2",
    "L-1x2": "x 4",
    "shaft-10cm": "",
    "M3": "",
    "front-connector": "",
    "front-wheels": "x 2",
}



from bluer_ugv.README.arzhang.items import items
from bluer_ugv.README.arzhang import design, algo, flag
from bluer_ugv.README.arzhang.validations import docs as validations

docs = (
    [
        {
            "items": items,
            "path": "../docs/arzhang",
        }
    ]
    + algo.docs
    + design.docs
    + flag.docs
    + validations.docs
)



from RPi import GPIO  # type: ignore

from bluer_options import string
from bluer_options.timing.classes import Timing
from bluer_objects.env import abcli_object_name
from bluer_objects.metadata import post_to_object
from bluer_sbc.env import BLUER_SBC_ENV, BLUER_SBC_SWALLOW_HAS_STEERING

from bluer_ugv.swallow.session.classical.camera import (
    ClassicalCamera,
    ClassicalNavigationCamera,
    ClassicalTrackingCamera,
    ClassicalYoloCamera,
)
from bluer_ugv.swallow.session.classical.push_button import ClassicalPushButton
from bluer_ugv.swallow.session.classical.keyboard import ClassicalKeyboard
from bluer_ugv.swallow.session.classical.leds import ClassicalLeds
from bluer_ugv.swallow.session.classical.mousepad import ClassicalMousePad
from bluer_ugv.swallow.session.classical.motor import (
    ClassicalLeftMotor,
    ClassicalRightMotor,
    ClassicalRearMotors,
    ClassicalSteeringMotor,
)
from bluer_ugv.swallow.session.classical.setpoint.classes import ClassicalSetPoint
from bluer_ugv.swallow.session.classical.ultrasonic_sensor.classes import (
    ClassicalUltrasonicSensor,
)
from bluer_ugv.env import BLUER_UGV_MOUSEPAD_ENABLED
from bluer_ugv.logger import logger


class ClassicalSession:
    def __init__(
        self,
        object_name: str,
    ):
        self.object_name = object_name

        GPIO.setmode(GPIO.BCM)

        self.leds = ClassicalLeds()

        self.setpoint = ClassicalSetPoint(
            leds=self.leds,
        )

        if BLUER_UGV_MOUSEPAD_ENABLED:
            self.mousepad = ClassicalMousePad(
                leds=self.leds,
                setpoint=self.setpoint,
            )

        self.keyboard = ClassicalKeyboard(
            leds=self.leds,
            setpoint=self.setpoint,
        )

        self.ultrasonic_sensor = ClassicalUltrasonicSensor(
            setpoint=self.setpoint,
            keyboard=self.keyboard,
        )

        self.push_button = ClassicalPushButton(
            leds=self.leds,
        )

        self.has_steering = BLUER_SBC_SWALLOW_HAS_STEERING == 1
        logger.info("has_steering: {}".format(self.has_steering))

        self.motor1 = (
            ClassicalSteeringMotor if self.has_steering else ClassicalRightMotor
        )(
            setpoint=self.setpoint,
            leds=self.leds,
        )

        self.motor2 = (
            ClassicalRearMotors if self.has_steering else ClassicalLeftMotor
        )(
            setpoint=self.setpoint,
            leds=self.leds,
        )

        logger.info(
            "wheel arrangement: {} + {}".format(
                self.motor1.role,
                self.motor2.role,
            )
        )

        camera_class = (
            ClassicalYoloCamera
            if BLUER_SBC_ENV == "yolo"
            else (
                ClassicalTrackingCamera
                if BLUER_SBC_ENV == "tracking"
                else (
                    ClassicalNavigationCamera
                    if BLUER_SBC_ENV == "navigation"
                    else ClassicalCamera
                )
            )
        )
        logger.info(f"camera: {camera_class.__name__}")
        self.camera = camera_class(
            keyboard=self.keyboard,
            leds=self.leds,
            setpoint=self.setpoint,
            object_name=self.object_name,
        )

        logger.info(
            "{}: created for {}".format(
                self.__class__.__name__,
                self.object_name,
            )
        )

        self.timing = Timing()

    def cleanup(self):
        self.ultrasonic_sensor.stop()
        self.camera.stop()

        for thing in [
            self.motor1,
            self.motor2,
            self.camera,
        ]:
            thing.cleanup()

        GPIO.cleanup()

        self.timing.calculate()
        loop_frequency = round(
            1 / self.timing.stats["session.update"]["average"],
            2,
        )
        self.timing.log()
        post_to_object(
            abcli_object_name,
            "timing",
            self.timing.as_dict,
        )
        logger.info(
            "loop frequency: {}".format(string.pretty_frequency(loop_frequency))
        )
        post_to_object(
            abcli_object_name,
            "loop_frequency",
            loop_frequency,
        )

        logger.info(f"{self.__class__.__name__}.cleanup")

    def initialize(self) -> bool:
        return all(
            thing.initialize()
            for thing in [
                self.push_button,
                self.leds,
                self.motor1,
                self.motor2,
                self.camera,
            ]
        )

    def update(self) -> bool:
        self.timing.start("session.update")

        for thing in [
            self.keyboard,
            self.push_button,
            self.camera,
            self.setpoint,
            self.motor1,
            self.motor2,
            self.leds,
        ]:
            self.timing.start(thing.__class__.__name__)

            if not thing.update():
                return False

            self.timing.stop(thing.__class__.__name__)

        self.timing.stop("session.update")
        return True



from bluer_ai.help.generic import help_functions as generic_help_functions

from bluer_ugv import ALIAS
from bluer_ugv.help.git import help_git
from bluer_ugv.help.swallow import help_functions as help_swallow


help_functions = generic_help_functions(plugin_name=ALIAS)

help_functions.update(
    {
        "git": help_git,
        "swallow": help_swallow,
    }
)



import keyboard
import threading
from typing import Any, Dict

from bluer_sbc.session.functions import reply_to_bash
from bluer_algo.socket.classes import DEV_HOST

from bluer_ugv.swallow.session.classical.setpoint.classes import ClassicalSetPoint
from bluer_ugv.swallow.session.classical.mode import OperationMode
from bluer_ugv.swallow.session.classical.leds import ClassicalLeds
from bluer_ugv import env
from bluer_ugv.logger import logger

bash_keys = {
    "i": "exit",
    "o": "shutdown",
    "p": "reboot",
    "u": "update",
}


class ClassicalKeyboard:
    def __init__(
        self,
        leds: ClassicalLeds,
        setpoint: ClassicalSetPoint,
    ):
        logger.info(
            "{}: {}".format(
                self.__class__.__name__,
                ", ".join(
                    [f"{key}:{action}" for key, action in bash_keys.items()],
                ),
            )
        )

        self.leds = leds

        self.last_key: str = ""

        self.setpoint = setpoint

        self.special_key: bool = False

        self._lock = threading.Lock()
        self.config: Dict[str, Any] = {
            "debug_mode": False,
            "mode": OperationMode.NONE,
            "ultrasound_enabled": True,
        }

    def get(self, what: str, default: Any) -> Any:
        with self._lock:
            return self.config.get(what, default)

    def set(self, what: str, value: Any):
        with self._lock:
            self.config[what] = value

    def update(self) -> bool:
        self.last_key = ""

        # bash keys
        if self.special_key:
            for key, event in bash_keys.items():
                if keyboard.is_pressed(key):
                    reply_to_bash(event)
                    return False

        # other keys
        for key, func in {
            " ": self.setpoint.stop,
            "x": self.setpoint.start,
            "s": lambda: self.setpoint.put(
                what="speed",
                value=self.setpoint.get(what="speed") - 10,
            ),
            "w": lambda: self.setpoint.put(
                what="speed",
                value=self.setpoint.get(what="speed") + 10,
            ),
        }.items():
            if keyboard.is_pressed(key):
                self.special_key = False
                func()

        # steering
        if keyboard.is_pressed("a"):
            self.special_key = False
            self.last_key = "a"
            self.setpoint.put(
                what="steering",
                value=env.BLUER_UGV_SWALLOW_STEERING_SETPOINT,
            )
        elif keyboard.is_pressed("d"):
            self.special_key = False
            self.last_key = "d"
            self.setpoint.put(
                what="steering",
                value=-env.BLUER_UGV_SWALLOW_STEERING_SETPOINT,
            )
        else:
            self.setpoint.put(
                what="steering",
                value=0,
                log=False,
            )

        # debug mode
        if keyboard.is_pressed("b"):
            self.special_key = False
            self.set("debug_mode", True)
            logger.info(f'debug enabled, run "@swallow debug" on {DEV_HOST}.')

        if keyboard.is_pressed("v"):
            self.special_key = False
            self.set("debug_mode", False)
            logger.info("debug disabled.")

        # mode
        mode = self.get("mode", OperationMode.NONE)
        updated_mode = mode
        if keyboard.is_pressed("y"):
            updated_mode = OperationMode.NONE

        if keyboard.is_pressed("t"):
            updated_mode = OperationMode.TRAINING

        if keyboard.is_pressed("g"):
            updated_mode = OperationMode.ACTION

        if mode != updated_mode:
            self.set("mode", updated_mode)
            logger.info("mode: {}.".format(updated_mode.name.lower()))
            self.special_key = False

        # ultrasound
        if keyboard.is_pressed("n"):
            self.set("ultrasound_enabled", False)
            logger.info("ultrasound: disabled")
            self.special_key = False

        if keyboard.is_pressed("m"):
            self.set("ultrasound_enabled", True)
            logger.info("ultrasound: enabled")
            self.special_key = False

        # special key
        if keyboard.is_pressed("z") and not self.special_key:
            self.special_key = True
            logger.info("🪄 special key enabled.")

        if self.special_key:
            self.leds.flash_all()

        return True



import threading
import time

from bluer_objects.env import abcli_object_name

from bluer_ugv import env
from bluer_ugv.swallow.session.classical.ultrasonic_sensor.pack import (
    UltrasonicSensorPack,
)
from bluer_ugv.swallow.session.classical.ultrasonic_sensor.detection import (
    DetectionState,
)
from bluer_ugv.swallow.session.classical.ultrasonic_sensor.log import (
    UltrasonicSensorDetectionLog,
)
from bluer_ugv.swallow.session.classical.keyboard import ClassicalKeyboard
from bluer_ugv.swallow.session.classical.setpoint.classes import ClassicalSetPoint
from bluer_ugv.logger import logger


class ClassicalUltrasonicSensor:
    def __init__(
        self,
        setpoint: ClassicalSetPoint,
        keyboard: ClassicalKeyboard,
    ):
        self.enabled = env.BLUER_UGV_ULTRASONIC_SENSOR_ENABLED == 1
        logger.info(
            "{}: {}".format(
                self.__class__.__name__,
                (
                    "enabled: warning<{:.2f} mm, danger<{:.2f} mm".format(
                        env.BLUER_UGV_ULTRASONIC_SENSOR_WARNING_THRESHOLD,
                        env.BLUER_UGV_ULTRASONIC_SENSOR_DANGER_THRESHOLD,
                    )
                    if self.enabled
                    else "disabled"
                ),
            )
        )

        self.setpoint = setpoint
        self.keyboard = keyboard

        self.pack = None
        self.log = None
        self.running = False

        if not self.enabled:
            return

        self.pack = UltrasonicSensorPack(
            setmode=False,
            max_m=env.BLUER_UGV_ULTRASONIC_SENSOR_MAX_M,
        )

        if not self.pack.valid:
            raise NameError("valid ultrasonic sensor pack.")

        if env.BLUER_UGV_ULTRASONIC_SENSOR_KEEP_LOG == 1:
            self.log = UltrasonicSensorDetectionLog()

        self.running = True
        self.thread = threading.Thread(target=self.loop, daemon=True)
        self.thread.start()

    def stop(self):
        if not self.enabled:
            return

        self.running = False
        self.thread.join()

        logger.info(f"{self.__class__.__name__}.stopped.")

        if self.log is not None:
            self.log.save(object_name=abcli_object_name)
            self.log.export(object_name=abcli_object_name)

    def loop(self):
        logger.info(f"{self.__class__.__name__}.loop started.")

        while self.running:
            if not self.keyboard.get("ultrasound_enabled", True):
                time.sleep(0.01)
                continue

            success, detections = self.pack.detect(
                log=env.BLUER_UGV_ULTRASONIC_SENSOR_LOG == 1
            )
            if not success:
                raise NameError("failed to detect ultrasonic sensor.")

            if self.log is not None:
                self.log.append(detections)

            log_detections: bool = False
            speed = self.setpoint.get(what="speed")
            if any(
                detection.state == DetectionState.DANGER for detection in detections
            ):
                self.setpoint.stop()
                log_detections = True
                logger.info("⛔️ danger detected, stopping.")
            elif (
                any(
                    detection.state == DetectionState.WARNING
                    for detection in detections
                )
                and speed > 0
            ):
                self.setpoint.put(
                    what="speed",
                    value=speed // 2,
                )
                log_detections = True
                logger.info("⚠️ warning detected, lowering speed.")

            if env.BLUER_UGV_ULTRASONIC_SENSOR_LOG == 1:
                log_detections = False

            if log_detections:
                logger.info(
                    "{}: {}".format(
                        self.__class__.__name__,
                        ", ".join(
                            [detection.as_str(short=True) for detection in detections]
                        ),
                    )
                )
